from applitools.common import (
    DeviceName,
    FileLogger,
    MatchLevel,
    RectangleSize,
    Region,
    StdoutLogger,
    TestResultContainer,
    TestResults,
    TestResultsSummary,
    logger,
)
from applitools.common.accessibility import (  # noqa
    AccessibilityGuidelinesVersion,
    AccessibilityLevel,
    AccessibilityRegionType,
    AccessibilitySettings,
)
from applitools.common.config import BatchInfo  # noqa
from applitools.common.geometry import AccessibilityRegion
from applitools.common.selenium import BrowserType, Configuration, StitchMode  # noqa
from applitools.common.server import FailureReports  # noqa
from applitools.common.ultrafastgrid import (  # noqa
    ChromeEmulationInfo,
    DesktopBrowserInfo,
    IosDeviceInfo,
    IosDeviceName,
    IosVersion,
    ScreenOrientation,
)
from applitools.core.batch_close import BatchClose  # noqa
from applitools.core.cut import (  # noqa
    FixedCutProvider,
    NullCutProvider,
    UnscaledFixedCutProvider,
)
from applitools.core.fluent.region import AccessibilityRegionByRectangle  # noqa

from .classic_runner import ClassicRunner  # noqa
from .extract_text import OCRRegion, TextRegionSettings
from .eyes import Eyes  # noqa
from .fluent.target import Target  # noqa
from .visual_grid import RunnerOptions, VisualGridRunner  # noqa
from .webdriver import EyesWebDriver  # noqa
from .webelement import EyesWebElement  # noqa

__all__ = (
    # noqa
    "BatchInfo",
    "Region",
    "MatchLevel",
    "logger",
    "StdoutLogger",
    "FileLogger",
    "Eyes",
    "Target",
    "FailureReports",
    "StitchMode",
    "VisualGridRunner",
    "RunnerOptions",
    "BrowserType",
    "DeviceName",
    "Configuration",
    "ScreenOrientation",
    "FixedCutProvider",
    "NullCutProvider",
    "UnscaledFixedCutProvider",
    "ClassicRunner",
    "RectangleSize",
    "TestResults",
    "TestResultContainer",
    "TestResultsSummary",
    "BatchClose",
    "AccessibilityRegionType",
    "AccessibilityLevel",
    "AccessibilitySettings",
    "AccessibilityGuidelinesVersion",
    "AccessibilityRegionByRectangle",
    "AccessibilityRegion",
    "IosDeviceName",
    "IosDeviceInfo",
    "IosVersion",
    "ChromeEmulationInfo",
    "DesktopBrowserInfo",
    "OCRRegion",
    "TextRegionSettings",
)



import typing
import uuid

import attr

from applitools.common import TestResults, logger

if typing.TYPE_CHECKING:
    from typing import Callable, Optional, Text

    from structlog import BoundLogger


@attr.s(hash=True)
class VGTask(object):
    name = attr.ib()  # type: Text
    func_to_run = attr.ib(hash=False, repr=False)  # type: Callable
    logger = attr.ib(hash=False, repr=False)  # type: BoundLogger
    uuid = attr.ib(init=False, repr=False, factory=lambda: str(uuid.uuid4()))

    callback = None
    error_callback = None
    complete_callback = None

    def __attrs_post_init__(self):
        self.logger = self.logger.bind(task_name=self.name)

    def on_task_succeeded(self, code):
        # type: (Callable) -> VGTask
        self.callback = code
        return self

    def on_task_error(self, code):
        # type: (Callable) -> VGTask
        self.error_callback = code
        return self

    def on_task_completed(self, code):
        # type: (Callable) -> VGTask
        self.complete_callback = code
        return self

    def __call__(self):
        # type: () -> Optional[TestResults]
        self.logger.debug("Executing " + self.__class__.__name__)
        res = None
        try:
            if callable(self.func_to_run):
                self.logger.debug(
                    "VGTask().func_to_run", func_to_run=self.func_to_run.__name__
                )
                res = self.func_to_run()
            if callable(self.callback):
                self.logger.debug("VGTask().callback", callback=self.callback.__name__)
                self.callback(res)
        except Exception as e:
            self.logger.error("Failed to execute task!", exc_info=e)
            if callable(self.error_callback):
                self.logger.debug(
                    "VGTask().error_callback",
                    error_callback=self.error_callback.__name__,
                )
                self.error_callback(e)
        finally:
            if callable(self.complete_callback):
                self.logger.debug(
                    "VGTask().complete_callback",
                    complete_callback=self.complete_callback.__name__,
                )
                self.complete_callback()
        return res



from typing import TYPE_CHECKING

from applitools.common import logger
from applitools.common.utils import argument_guard, image_utils
from applitools.core.locators import (
    LOCATORS_TYPE,
    VisualLocatorsData,
    VisualLocatorSettings,
    VisualLocatorsProvider,
)

if TYPE_CHECKING:
    from applitools.common.utils.custom_types import AnyWebDriver
    from applitools.core import ServerConnector
    from applitools.core.debug import DebugScreenshotsProvider
    from applitools.selenium.selenium_eyes import SeleniumEyes


class SeleniumVisualLocatorsProvider(VisualLocatorsProvider):
    def __init__(self, driver, eyes):
        # type: (AnyWebDriver, SeleniumEyes) -> None
        self._driver = driver
        self._eyes = eyes
        self._server_connector = eyes._server_connector  # type: ServerConnector
        self._debug_screenshot_provider = (
            eyes.debug_screenshots_provider
        )  # type: DebugScreenshotsProvider

    def _get_viewport_screenshot(self):
        scale_provider = self._eyes.update_scaling_params()
        return self._eyes.get_scaled_cropped_viewport_image(scale_provider)

    def get_locators(self, visual_locator_settings):
        # type: (VisualLocatorSettings) -> LOCATORS_TYPE
        argument_guard.not_none(visual_locator_settings)

        logger.info(
            "Get locators with given names: {}".format(visual_locator_settings.names)
        )
        logger.info("Requested viewport screenshot for visual locators..")
        viewport_screenshot = self._get_viewport_screenshot()
        self._debug_screenshot_provider.save(
            viewport_screenshot,
            "Visual locators: {}".format(visual_locator_settings.names),
        )
        image = image_utils.get_bytes(viewport_screenshot)
        logger.info("Post visual locators screenshot...")
        viewport_screenshot_url = self._server_connector.try_upload_image(image)

        logger.info("Screenshot URL: {}".format(viewport_screenshot_url))
        data = VisualLocatorsData(
            app_name=self._eyes.configure.app_name,
            image_url=viewport_screenshot_url,
            first_only=visual_locator_settings.values.is_first_only,
            locator_names=visual_locator_settings.values.names,
        )
        logger.info("Post visual locators: {}".format(data))
        return self._server_connector.post_locators(data)



from __future__ import absolute_import

import typing

from appium.webdriver import Remote as AppiumWebDriver

from applitools.common import EyesError, MatchResult, deprecated, logger
from applitools.common.selenium import Configuration
from applitools.common.utils import argument_guard
from applitools.common.utils.compat import basestring
from applitools.common.utils.general_utils import all_fields, proxy_to
from applitools.core.eyes_base import DebugScreenshotsAbstract, ExtractTextMixin
from applitools.core.eyes_mixins import EyesConfigurationMixin
from applitools.core.feature import Feature
from applitools.core.locators import LOCATORS_TYPE, VisualLocatorSettings
from applitools.core.server_connector import ServerConnector
from applitools.selenium import ClassicRunner, eyes_selenium_utils

from .extract_text import (
    PATTERN_TEXT_REGIONS,
    OCRRegion,
    SeleniumExtractTextProvider,
    TextRegionSettings,
)
from .fluent import SeleniumCheckSettings, Target
from .locators import SeleniumVisualLocatorsProvider
from .selenium_eyes import SeleniumEyes
from .visual_grid import VisualGridEyes, VisualGridRunner
from .webdriver import EyesWebDriver

if typing.TYPE_CHECKING:
    from typing import List, Optional, Text, Tuple, Union

    from selenium.webdriver.remote.webelement import WebElement

    from applitools.common import MatchLevel, Region, TestResults
    from applitools.common.utils.custom_types import (
        AnyWebDriver,
        AnyWebElement,
        FrameReference,
        ViewPort,
    )
    from applitools.core import (
        FixedCutProvider,
        NullCutProvider,
        PositionProvider,
        UnscaledFixedCutProvider,
    )

    from .frames import FrameChain
    from .webelement import EyesWebElement


@proxy_to("configure", all_fields(Configuration))
class Eyes(EyesConfigurationMixin, DebugScreenshotsAbstract, ExtractTextMixin):
    _is_visual_grid_eyes = False  # type: bool
    _visual_grid_eyes = None  # type: VisualGridEyes
    _selenium_eyes = None  # type: SeleniumEyes
    _visual_locators_provider = None  # type: SeleniumVisualLocatorsProvider
    _runner = None  # type: Optional[VisualGridRunner]
    _driver = None  # type: Optional[EyesWebDriver]
    _is_opened = False  # type: bool
    _config_cls = Configuration
    _extract_text_provider = None  # type: Optional[SeleniumExtractTextProvider]

    def __init__(self, runner=None):
        # type: (Union[Text, VisualGridRunner, ClassicRunner, None]) -> None
        super(Eyes, self).__init__()

        # backward compatibility with settings server_url
        if runner is None:
            self._selenium_eyes = SeleniumEyes(self, None)
        elif isinstance(runner, str):
            self.configure.server_url = runner
            self._selenium_eyes = SeleniumEyes(self, None)
        elif isinstance(runner, VisualGridRunner):
            self._runner = runner
            self._visual_grid_eyes = VisualGridEyes(self, runner)
            self._is_visual_grid_eyes = True
            # for visual locators
            self._selenium_eyes = SeleniumEyes(self, None)
        elif isinstance(runner, ClassicRunner):
            self._runner = runner
            self._selenium_eyes = SeleniumEyes(self, runner)
            self._is_visual_grid_eyes = False
        else:
            raise TypeError(
                "{} is unsupported runner. Must be {} or {}".format(
                    type(runner), ClassicRunner.__name__, VisualGridRunner.__name__
                )
            )

    def get_configuration(self):
        # type:() -> Configuration
        return super(Eyes, self).get_configuration()

    def set_configuration(self, configuration):
        # type:(Configuration) -> None
        return super(Eyes, self).set_configuration(configuration)

    @property
    def server_connector(self):
        # type: () -> ServerConnector
        return self._current_eyes.server_connector

    @server_connector.setter
    def server_connector(self, server_connector):
        # type: (ServerConnector) -> None
        argument_guard.is_a(server_connector, ServerConnector)
        self._current_eyes.server_connector = server_connector

    @property
    def configure(self):
        # type:() -> Configuration
        return super(Eyes, self).configure

    @property
    def is_open(self):
        # type: () -> bool
        return self._is_opened

    @property
    def rotation(self):
        # type: () -> Optional[int]
        if self._selenium_eyes and self.driver:
            return self.driver.rotation
        return None

    @rotation.setter
    def rotation(self, rotation):
        # type: (Union[int,float]) -> None
        argument_guard.is_in(rotation, [int, float])
        if self._selenium_eyes and self.driver:
            self.driver.rotation = int(rotation)

    @property
    def base_agent_id(self):
        # type: () -> Text
        """
        Must return version of SDK. (e.g. selenium, visualgrid) in next format:
            "eyes.{package}.python/{lib_version}"
        """
        return self._current_eyes.base_agent_id

    @property
    def full_agent_id(self):
        # type: () -> Text
        """
        Gets the agent id, which identifies the current library using the SDK.

        """
        return self._current_eyes.full_agent_id

    @property
    def match_level(self):
        # type: () -> MatchLevel
        return self.configure.match_level

    @match_level.setter
    def match_level(self, match_level):
        # type: (MatchLevel) -> None
        self.configure.match_level = match_level

    @property
    def should_stitch_content(self):
        # type: () -> bool
        return self._current_eyes.should_stitch_content

    @property
    def original_fc(self):
        # type: () -> Optional[FrameChain]
        """Gets original frame chain

        Before check() call we save original frame chain

        Returns:
            Frame chain saved before check() call
        """
        return self._current_eyes.original_fc

    @property
    def device_pixel_ratio(self):
        # type: () -> int
        """
        Gets device pixel ratio.

        :return The device pixel ratio, or if the DPR is not known yet or if it wasn't
        possible to extract it.
        """
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.device_pixel_ratio
        return 0

    @property
    def scale_ratio(self):
        # type: () -> float
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.scale_ratio
        return 0

    @scale_ratio.setter
    def scale_ratio(self, value):
        # type: (float) -> None
        """
        Manually set the scale ratio for the images being validated.
        """
        if not self._is_visual_grid_eyes:
            self._selenium_eyes.scale_ratio = value

    @property
    def position_provider(self):
        """
        Sets position provider.
        """
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.position_provider
        return None

    @property
    def save_debug_screenshots(self):
        # type: () -> bool
        """True if screenshots saving enabled."""
        return self._selenium_eyes.save_debug_screenshots

    @save_debug_screenshots.setter
    def save_debug_screenshots(self, save):
        # type: (bool) -> None
        """If True, will save all screenshots to local directory."""
        self._selenium_eyes.save_debug_screenshots = save

    @property
    def debug_screenshots_provider(self):
        return self._selenium_eyes.debug_screenshots_provider

    @property
    def debug_screenshots_path(self):
        # type: () -> Optional[Text]
        """The path where you save the debug screenshots."""
        return self._selenium_eyes.debug_screenshots_path

    @debug_screenshots_path.setter
    def debug_screenshots_path(self, path_to_save):
        # type: (Text) -> None
        """The path where you want to save the debug screenshots."""
        self._selenium_eyes.debug_screenshots_path = path_to_save

    @property
    def debug_screenshots_prefix(self):
        # type: () -> Optional[Text]
        """The prefix for the screenshots' names."""
        return self._selenium_eyes.debug_screenshots_prefix

    @debug_screenshots_prefix.setter
    def debug_screenshots_prefix(self, prefix):
        # type: (Text) -> None
        """The prefix for the screenshots' names."""
        self._selenium_eyes.debug_screenshots_prefix = prefix

    @position_provider.setter
    def position_provider(self, provider):
        # type: (PositionProvider) -> None
        """
        Gets position provider.
        """
        self._selenium_eyes.position_provider = provider

    @property
    def cut_provider(self):
        # type:()->Union[FixedCutProvider,UnscaledFixedCutProvider,NullCutProvider,None]
        """
        Gets current cut provider
        """
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.cut_provider
        return None

    @cut_provider.setter
    def cut_provider(self, cutprovider):
        # type: (Union[FixedCutProvider,UnscaledFixedCutProvider,NullCutProvider])->None
        """
        Manually set the the sizes to cut from an image before it's validated.
        """
        if not self._is_visual_grid_eyes:
            self._selenium_eyes.cut_provider = cutprovider

    @property
    def is_cut_provider_explicitly_set(self):
        # type: () -> bool
        """
        Gets is cut provider explicitly set.
        """
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.is_cut_provider_explicitly_set
        return False

    @property
    def agent_setup(self):
        # type: () -> Optional[Text]
        """
        Gets agent setup.
        """
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.agent_setup
        return None

    @property
    def current_frame_position_provider(self):
        # type: () -> Optional[PositionProvider]
        if not self._is_visual_grid_eyes:
            return self._selenium_eyes.current_frame_position_provider
        return None

    @staticmethod
    def get_viewport_size(driver):
        # type: (AnyWebDriver) -> ViewPort
        argument_guard.not_none(driver)
        return eyes_selenium_utils.get_viewport_size_or_display_size(driver)

    @staticmethod
    def set_viewport_size(driver, size):
        # type: (AnyWebDriver, ViewPort) -> None
        argument_guard.not_none(driver)
        argument_guard.not_none(size)
        eyes_selenium_utils.set_viewport_size(driver, size)

    def add_property(self, name, value):
        # type: (Text, Text) -> None
        """
        Associates a key/value pair with the test. This can be used later for filtering.
        :param name: (string) The property name.
        :param value: (string) The property value
        """
        self._current_eyes.configure.add_property(name, value)

    def clear_properties(self):
        """
        Clears the list of custom properties.
        """
        self._current_eyes.configure.clear_properties()

    def add_mouse_trigger_by_element(self, action, element):
        # type: (Text, AnyWebElement) -> None
        """
        Adds a mouse trigger.

        :param action: Mouse action (click, double click etc.)
        :param element: The element on which the action was performed.
        """
        if self.configure.is_disabled:
            logger.debug("add_mouse_trigger: Ignoring %s (disabled)" % action)
            return
        if not self._is_visual_grid_eyes:
            self._selenium_eyes.add_mouse_trigger_by_element(action, element)

    def add_text_trigger_by_element(self, element, text):
        # type: (AnyWebElement, Text) -> None
        """
        Adds a text trigger.

        :param element: The element to which the text was sent.
        :param text: The trigger's text.
        """
        if self.configure.is_disabled:
            logger.debug("add_text_trigger: Ignoring '%s' (disabled)" % text)
            return
        if not self._is_visual_grid_eyes:
            self._selenium_eyes.add_text_trigger_by_element(element, text)

    @property
    def driver(self):
        # type: () -> EyesWebDriver
        return self._driver

    @property
    def send_dom(self):
        # type: () -> bool
        if not self._is_visual_grid_eyes:
            return self.configure.send_dom
        return False

    @send_dom.setter
    def send_dom(self, value):
        # type: (bool) -> None
        if not self._is_visual_grid_eyes:
            self.configure.send_dom = value

    @typing.overload
    def check(self, name, check_settings):
        # type: (Text, SeleniumCheckSettings) -> MatchResult
        """
        Takes a snapshot and matches it with the expected output.

        :param name: The name of the tag.
        :param check_settings: target which area of the window to check.
        """
        pass

    @typing.overload
    def check(self, check_settings):
        # type: (SeleniumCheckSettings) -> None
        """
        Takes a snapshot and matches it with the expected output.

        :param check_settings: target which area of the window to check.
        """
        pass

    def check(self, check_settings, name=None):
        if isinstance(name, SeleniumCheckSettings) or isinstance(
            check_settings, basestring
        ):
            check_settings, name = name, check_settings
        if check_settings is None:
            check_settings = Target.window()
        if name:
            check_settings = check_settings.with_name(name)

        if self.configure.is_disabled:
            logger.info("check(): ignored (disabled)")
            return MatchResult()
        if not self.is_open:
            self.abort()
            raise EyesError("you must call open() before checking")
        return self._current_eyes.check(check_settings)

    def check_window(self, tag=None, match_timeout=-1, fully=None):
        # type: (Optional[Text], int, Optional[bool]) -> MatchResult
        """
        Takes a snapshot of the application under test and matches it with the expected
         output.

        :param tag: An optional tag to be associated with the snapshot.
        :param match_timeout:  The amount of time to retry matching (milliseconds)
        :param fully: Defines that the screenshot will contain the entire window.
        """
        logger.debug("check_window('%s')" % tag)
        return self.check(tag, Target.window().timeout(match_timeout).fully(fully))

    def check_frame(self, frame_reference, tag=None, match_timeout=-1):
        # type: (FrameReference, Optional[Text], int) -> MatchResult
        """
        Check frame.

        :param frame_reference: The name or id of the frame to check. (The same
                name/id as would be used in a call to driver.switch_to.frame()).
        :param tag: An optional tag to be associated with the match.
        :param match_timeout: The amount of time to retry matching. (Milliseconds)
        """
        logger.debug("check_frame('%s')" % tag)
        return self.check(
            tag, Target.frame(frame_reference).fully().timeout(match_timeout)
        )

    def check_region(
        self,
        region,  # type: Union[Region,Text,List,Tuple,WebElement,EyesWebElement]
        tag=None,  # type: Optional[Text]
        match_timeout=-1,  # type: int
        stitch_content=False,  # type: bool
    ):
        # type: (...) -> MatchResult
        """
        Takes a snapshot of the given region from the browser using the web driver
        and matches it with the expected output. If the current context is a frame,
        the region is offsetted relative to the frame.

        :param region: The region which will be visually validated. The coordinates are
                       relative to the viewport of the current frame.
        :param tag: Description of the visual validation checkpoint.
        :param match_timeout: Timeout for the visual validation checkpoint
                              (milliseconds).
        :param stitch_content: If `True`, stitch the internal content of the region
        """
        logger.debug("check_region('%s')" % tag)
        return self.check(
            tag,
            Target.region(region).timeout(match_timeout).fully(stitch_content),
        )

    def check_element(
        self,
        element,  # type: Union[Text,List,Tuple,WebElement,EyesWebElement]
        tag=None,  # type: Optional[Text]
        match_timeout=-1,  # type: int
    ):
        # type: (...) -> MatchResult
        """
        Takes a snapshot of the given region from the browser using the web driver
        and matches it with the expected output. If the current context is a frame,
        the region is offsetted relative to the frame.

        :param element: The element to check.
        :param tag: Description of the visual validation checkpoint.
        :param match_timeout: Timeout for the visual validation checkpoint
                              (milliseconds).
        """
        logger.debug("check_region('%s')" % tag)
        return self.check(
            tag,
            Target.region(element).timeout(match_timeout).fully(),
        )

    def check_region_in_frame(
        self,
        frame_reference,  # type: FrameReference
        region,  # type: Union[Region,Text,List,Tuple,WebElement,EyesWebElement]
        tag=None,  # type: Optional[Text]
        match_timeout=-1,  # type: int
        stitch_content=False,  # type: bool
    ):
        # type: (...) -> MatchResult
        """
        Checks a region within a frame, and returns to the current frame.

        :param frame_reference: A reference to the frame in which the region
                                should be checked.
        :param region: Specifying the region to check inside the frame.
        :param tag: Description of the visual validation checkpoint.
        :param match_timeout: Timeout for the visual validation checkpoint
                              (milliseconds).
        :param stitch_content: If `True`, stitch the internal content of the region
        """
        if self.configure.is_disabled:
            logger.info("check_region_in_frame_by_selector(): ignored (disabled)")
            return MatchResult()
        logger.debug("check_region_in_frame_by_selector('%s')" % tag)
        return self.check(
            tag,
            Target.region(region)
            .frame(frame_reference)
            .stitch_content(stitch_content)
            .timeout(match_timeout),
        )

    def open(
        self,
        driver,  # type: AnyWebDriver
        app_name=None,  # type: Optional[Text]
        test_name=None,  # type: Optional[Text]
        viewport_size=None,  # type: Optional[ViewPort]
    ):
        # type: (...) -> Optional[EyesWebDriver]
        """
        Starts a test.

        :param driver: The driver that controls the browser hosting the application
            under the test.
        :param app_name: The name of the application under test.
        :param test_name: The test name.
        :param viewport_size: The client's viewport size (i.e.,
            the visible part of the document's body) or None to allow any viewport size.
        :raise EyesError: If the session was already open.
        """
        if self.configure.is_disabled:
            logger.info("open(): ignored (disabled)")
            return

        if app_name:
            self.configure.app_name = app_name
        if test_name:
            self.configure.test_name = test_name
        if viewport_size:
            self.configure.viewport_size = viewport_size  # type: ignore

        argument_guard.not_none(
            self.configure.app_name, ValueError("app_name is required")
        )
        argument_guard.not_none(
            self.configure.test_name, ValueError("test_name is required")
        )

        logger.info(
            "open(app_name={}, test_name={}, viewport_size={})".format(
                self.configure.app_name,
                self.configure.test_name,
                self.configure.viewport_size,
            )
        )

        self._init_driver(driver)
        self._init_additional_providers()
        if self._is_visual_grid_eyes:
            self._selenium_eyes.open(self._driver, skip_start_session=True)

        result = self._current_eyes.open(self.driver)
        self._is_opened = True
        return result

    def locate(self, visual_locator_settings):
        # type: (VisualLocatorSettings) -> LOCATORS_TYPE
        argument_guard.is_a(visual_locator_settings, VisualLocatorSettings)
        logger.info("locate({})".format(visual_locator_settings))
        return self._visual_locators_provider.get_locators(visual_locator_settings)

    def extract_text(self, *regions):
        # type: (*OCRRegion) -> List[Text]
        return super(Eyes, self).extract_text(*regions)

    def extract_text_regions(self, config):
        # type: (TextRegionSettings) -> PATTERN_TEXT_REGIONS
        return super(Eyes, self).extract_text_regions(config)

    def close(self, raise_ex=True):
        # type: (bool) -> Optional[TestResults]
        """
        Ends the test.

        :param raise_ex: If true, an exception will be raised for failed/new tests.
        :return: The test results.
        """
        if self.configure.is_disabled:
            logger.info("close(): ignored (disabled)")
            return
        logger.info("close()")
        result = self._current_eyes.close(raise_ex)
        self._is_opened = False
        return result

    def close_async(self):
        if self.configure.is_disabled:
            logger.info("close_async(): ignored (disabled)")
            return
        logger.info("close_async()")
        if self._is_visual_grid_eyes:
            self._visual_grid_eyes.close_async()
        else:
            self._selenium_eyes.close(False)

    def abort(self):
        """
        If a test is running, aborts it. Otherwise, does nothing.
        """
        if self.configure.is_disabled:
            logger.info("abort(): ignored (disabled)")
            return
        logger.info("abort()")
        self._current_eyes.abort()

    def abort_async(self):
        if self.configure.is_disabled:
            logger.info("abort_async(): ignored (disabled)")
            return

        logger.info("abort_async()")
        if self._is_visual_grid_eyes:
            return self._visual_grid_eyes.abort_async()
        else:
            return self._selenium_eyes.abort()

    @deprecated.attribute("use `abort()` instead")
    def abort_if_not_closed(self):
        self.abort()

    def _init_driver(self, driver):
        # type: (AnyWebDriver) -> None
        if isinstance(driver, EyesWebDriver):
            # If the driver is an EyesWebDriver (as might be the case when tests are ran
            # consecutively using the same driver object)
            self._driver = driver
        else:
            self._driver = EyesWebDriver(driver, self)

        if self._driver.is_mobile_app and not isinstance(driver, AppiumWebDriver):
            logger.error("To test a mobile app you need to use appium webdriver")
            if Feature.SCALE_MOBILE_APP in self.configure.features:
                raise EyesError(
                    "For mobile app testing the appium driver should be used"
                )

    def _init_additional_providers(self):
        self._visual_locators_provider = SeleniumVisualLocatorsProvider(
            self._driver, self._selenium_eyes
        )
        self._extract_text_provider = SeleniumExtractTextProvider(
            self._driver, self._selenium_eyes
        )

    @property
    def _current_eyes(self):
        # type: () -> Union[SeleniumEyes, VisualGridEyes]
        if self._is_visual_grid_eyes:
            return self._visual_grid_eyes
        else:
            return self._selenium_eyes

    @property
    def _original_scroll_position(self):
        if self._selenium_eyes:
            return self._selenium_eyes._position_memento.position
        return None



import itertools
import typing
import uuid
from bisect import bisect_right
from copy import deepcopy

import attr

from applitools.common import EyesError, RectangleSize, TestResults, deprecated
from applitools.common.config import DEFAULT_ALL_TEST_RESULTS_TIMEOUT
from applitools.common.ultrafastgrid import (
    DesktopBrowserInfo,
    JobInfo,
    RenderBrowserInfo,
    RenderInfo,
    RenderRequest,
    VisualGridSelector,
    device_sizes_db,
)
from applitools.common.utils import argument_guard
from applitools.common.utils.compat import raise_from
from applitools.common.utils.datetime_utils import sleep
from applitools.common.utils.general_utils import random_alphanum
from applitools.core import CheckSettings, GetRegion, ServerConnector
from applitools.selenium import eyes_selenium_utils
from applitools.selenium.__version__ import __version__
from applitools.selenium.fluent import SeleniumCheckSettings
from applitools.selenium.visual_grid import dom_snapshot_script

from .eyes_connector import EyesConnector
from .helpers import collect_test_results, wait_till_tests_completed
from .resource_collection_task import ResourceCollectionTask
from .running_test import TESTED, RunningTest
from .visual_grid_runner import VisualGridRunner

if typing.TYPE_CHECKING:
    from typing import Dict, Generator, List, Optional, Text, Union

    from applitools.common.utils.custom_types import AnyWebElement
    from applitools.selenium import Configuration, Eyes, EyesWebDriver

    from .running_test import RunningTestCheck

GET_ELEMENT_XPATH_JS = """
var el = arguments[0];
var xpath = '';
do {
var parent = el.parentElement;
var index = 1;
if (parent !== null) {
var children = parent.children;
for (var childIdx in children) {
var child = children[childIdx];
if (child === el) break;
if (child.tagName === el.tagName) index++;
}
}
xpath = '/' + el.tagName + '[' + index + ']' + xpath;
el = parent;
} while (el !== null);
return '/' + xpath;"""

DOM_EXTRACTION_TIMEOUT = 10 * 60 * 1000
VIEWPORT_RESIZE_DELAY = 300


@attr.s
class WebElementRegion(object):
    region_provider = attr.ib()  # type: Union[Text, GetRegion]
    _webelement = attr.ib()  # type: AnyWebElement

    @property
    def webelement(self):
        return eyes_selenium_utils.get_underlying_webelement(self._webelement)


class VisualGridEyes(object):
    vg_manager = None  # type: VisualGridRunner
    _config_provider = None
    _is_opened = False
    _driver = None
    rendering_info = None
    is_check_timer_timeout = False

    def __init__(self, config_provider, runner):
        # type: (Eyes, VisualGridRunner)-> None
        self._config_provider = config_provider
        self._elements = []
        argument_guard.not_none(runner)
        self.vg_manager = runner  # type: VisualGridRunner
        self.test_list = []  # type: List[RunningTest]
        self._test_uuid = None
        self.server_connector = ServerConnector()  # type: ServerConnector
        self.resource_collection_queue = []  # type: List[ResourceCollectionTask]
        self.logger = runner.logger.bind(eyes_id=id(self))

    @property
    def is_open(self):
        return self._is_opened

    @property
    def driver(self):
        return self._driver

    @property
    def configure(self):
        # type: () -> Configuration
        return self._config_provider.configure

    @property
    def base_agent_id(self):
        # type: () -> Text
        """
        Must return version of SDK. (e.g. selenium, visualgrid) in next format:
            "eyes.{package}.python/{lib_version}"
        """
        return "eyes.selenium.visualgrid.python/{version}".format(version=__version__)

    @property
    def full_agent_id(self):
        # type: () -> Text
        """
        Gets the agent id, which identifies the current library using the SDK.

        """
        agent_id = self.configure.agent_id
        if agent_id is None:
            return self.base_agent_id
        return "{} [{}]".format(agent_id, self.base_agent_id)

    def _job_info_for_browser_info(self, browser_infos):
        # type: (List[RenderBrowserInfo]) -> Generator[RenderBrowserInfo, JobInfo]
        render_requests = [
            RenderRequest(
                render_info=RenderInfo.from_(
                    size_mode=None,
                    region=None,
                    selector=None,
                    render_browser_info=b_info,
                ),
                platform_name=b_info.platform,
                browser_name=b_info.browser,
            )
            for b_info in browser_infos
        ]

        job_infos = self.server_connector.job_info(render_requests)
        for b_info, job_info in zip(browser_infos, job_infos):
            yield b_info, job_info

    def open(self, driver):
        # type: (EyesWebDriver) -> EyesWebDriver
        self._test_uuid = uuid.uuid4()
        argument_guard.not_none(driver)
        self.logger.debug("VisualGridEyes.open()", config=self.configure)

        self._driver = driver
        self._set_viewport_size()
        self.server_connector.update_config(
            self.configure,
            self.full_agent_id,
            ua_string=self.driver.user_agent.origin_ua_string,
        )
        self.server_connector.render_info()
        self.vg_manager.rendering_service.maybe_set_server_connector(
            self.server_connector
        )
        device_sizes_db.maybe_set_server_connector(self.server_connector)

        for browser_info, job_info in self._job_info_for_browser_info(
            self.configure.browsers_info
        ):
            test = RunningTest(
                self._create_vgeyes_connector(browser_info, job_info),
                self.configure.clone(),
                browser_info,
                self.vg_manager.rendering_service,
                self.logger,
            )
            test.on_results_received(self.vg_manager.aggregate_result)
            test.test_uuid = self._test_uuid
            self.test_list.append(test)
            test.becomes_not_opened()
        self._is_opened = True
        self.vg_manager.open(self)
        self.logger.info("VisualGridEyes opening", tests_count=len(self.test_list))
        return driver

    def get_script_result(self, dont_fetch_resources):
        # type: (bool) -> Dict
        self.logger.debug(
            "get_script_result call", dont_fetch_resources=dont_fetch_resources
        )
        try:
            return dom_snapshot_script.create_dom_snapshot(
                self.driver,
                self.logger,
                dont_fetch_resources,
                None,
                DOM_EXTRACTION_TIMEOUT,
                self.configure.enable_cross_origin_rendering,
                not self.configure.dont_use_cookies,
            )
        except dom_snapshot_script.DomSnapshotFailure as e:
            raise_from(EyesError("Failed to capture dom snapshot"), e)

    @staticmethod
    def _options_dict(configuration_options, check_settings_options):
        return {
            o.key: o.value
            for o in itertools.chain(
                configuration_options or (), check_settings_options or ()
            )
        }

    def check(self, check_settings):
        # type: (SeleniumCheckSettings) -> None
        argument_guard.is_a(check_settings, CheckSettings)
        name = check_settings.values.name
        self.logger.debug(
            "VisualGridEyes.check", name=name, check_settings=check_settings
        )
        self._try_set_target_selector(check_settings)

        self.configure.send_dom = check_settings.values.send_dom

        check_settings = self._update_check_settings(check_settings)
        self.logger.info("check begin", name=name)

        region_xpaths = self.get_region_xpaths(check_settings)
        self.logger.info("check region xpaths", region_xpaths=region_xpaths)
        breakpoints = self._effective_layout_breakpoints(self.configure, check_settings)
        dont_fetch_resources = self._effective_disable_browser_fetching(
            self.configure, check_settings
        )
        running_tests = [
            test for test in self.test_list if self._test_uuid == test.test_uuid
        ]
        for width, tests in _group_tests_by_width(running_tests, breakpoints).items():
            try:
                self._capture_dom_and_schedule_resource_collection_and_checks(
                    width, check_settings, dont_fetch_resources, region_xpaths, tests
                )
            except Exception:
                self.logger.exception("Check failure")
                for test in tests:
                    if test.state != TESTED:
                        # already aborted or closed
                        test.abort()
                        test.becomes_tested()
            self.logger.info("added check tasks", check_settings=check_settings)
        self._set_viewport_size()

    def _capture_dom_and_schedule_resource_collection_and_checks(
        self, width, check_settings, dont_fetch_resources, region_xpaths, running_tests
    ):
        if width:
            try:
                requested = RectangleSize(width, self.configure.viewport_size.height)
                with self.driver.switch_to.frames_and_back([]):
                    eyes_selenium_utils.set_viewport_size(self.driver, requested)
                sleep(VIEWPORT_RESIZE_DELAY, "Waiting after viewport resize")
            except EyesError:
                actual = eyes_selenium_utils.get_viewport_size(self.driver)
                self.logger.warning(
                    "Failed to resize browser window. "
                    "Running browser in headless mode might prevent this error.",
                    requested=requested,
                    actual=actual,
                )
        script_result = self.get_script_result(dont_fetch_resources)
        self.logger.debug(
            "Got script result",
            width=width,
            cdt_len=len(script_result["cdt"]),
            blob_urls=[b["url"] for b in script_result["blobs"]],
            resource_urls=script_result["resourceUrls"],
        )
        source = eyes_selenium_utils.get_check_source(self.driver)
        checks = [
            test.check(
                check_settings=check_settings,
                region_selectors=region_xpaths,
                source=source,
            )
            for test in running_tests
        ]
        resource_collection_task = self._resource_collection_task(
            check_settings, region_xpaths, running_tests, script_result, checks
        )
        self.vg_manager.add_resource_collection_task(resource_collection_task)

    def _resource_collection_task(
        self,
        check_settings,  # type: SeleniumCheckSettings
        region_xpaths,  # type: List[List[VisualGridSelector]]
        running_tests,  # type: List[RunningTest]
        script_result,  # type: Dict
        checks,  # type: List[RunningTestCheck]
    ):
        # type: (...) -> ResourceCollectionTask
        tag = check_settings.values.name
        short_description = "{} of {}".format(
            self.configure.test_name, self.configure.app_name
        )
        resource_collection_task = ResourceCollectionTask(
            name="VisualGridEyes.check-resource_collection {} - {}".format(
                short_description, tag
            ),
            logger=self.logger,
            script=script_result,
            resource_cache=self.vg_manager.resource_cache,
            put_cache=self.vg_manager.put_cache,
            rendering_info=self.server_connector.render_info(),
            server_connector=self.server_connector,
            region_selectors=region_xpaths,
            size_mode=check_settings.values.size_mode,
            region_to_check=check_settings.values.target_region,
            script_hooks=check_settings.values.script_hooks,
            agent_id=self.base_agent_id,
            selector=check_settings.values.selector,
            running_tests=running_tests,
            request_options=self._options_dict(
                self.configure.visual_grid_options,
                check_settings.values.visual_grid_options,
            ),
        )

        def on_collected_task_succeeded(render_requests):
            for check in checks:
                check.set_render_request(render_requests[check.running_test])

        def on_collected_task_error(e):
            # TODO: Improve exception handling
            running_tests[0].pending_exceptions.append(e)
            for test in running_tests:
                if test.state != TESTED:
                    # already aborted or closed
                    test.abort()
                    test.becomes_tested()

        resource_collection_task.on_task_succeeded(on_collected_task_succeeded)
        resource_collection_task.on_task_error(on_collected_task_error)
        return resource_collection_task

    def close_async(self):
        for test in self.test_list:
            test.close()

    def close(self, raise_ex=True):  # noqa
        # type: (Optional[bool]) -> Optional[TestResults]
        if not self.test_list:
            return TestResults(status="Failed")
        self.logger.debug("VisualGridEyes.close()", test_list=self.test_list)
        self.close_async()

        wait_till_tests_completed(self.test_list, DEFAULT_ALL_TEST_RESULTS_TIMEOUT)

        self._is_opened = False

        all_results = collect_test_results(
            {t: t.test_result for t in self.test_list}, raise_ex
        )
        if not all_results:
            return TestResults(status="Failed")
        return all_results[0].test_results

    def abort_async(self):
        """
        If a test is running, aborts it. Otherwise, does nothing.
        """
        for test in self.test_list:
            if not test.is_closed:
                test.abort()

    def abort(self):
        """
        If a test is running, aborts it. Otherwise, does nothing.
        """
        self._is_opened = False
        self.abort_async()

    @deprecated.attribute("use `abort()` instead")
    def abort_if_not_closed(self):
        self.abort()

    def _update_check_settings(self, check_settings):
        # type: (SeleniumCheckSettings) -> SeleniumCheckSettings
        match_level = check_settings.values.match_level
        fully = check_settings.values.stitch_content
        send_dom = check_settings.values.send_dom
        ignore_displacements = check_settings.values.ignore_displacements

        if match_level is None:
            check_settings = check_settings.match_level(self.configure.match_level)
        if fully is None:
            fps = self.configure.force_full_page_screenshot
            check_settings = check_settings.fully(
                check_settings.is_check_window if fps is None else fps
            )
        if send_dom is None:
            send = self.configure.send_dom
            check_settings = check_settings.send_dom(True if send is None else send)
        if ignore_displacements is None:
            check_settings = check_settings.ignore_displacements(
                self.configure.ignore_displacements
            )
        return check_settings

    def _create_vgeyes_connector(self, b_info, job_info):
        # type: (RenderBrowserInfo, JobInfo) -> EyesConnector
        agent_run_id = "{}_{}".format(self.configure.test_name, random_alphanum(10))
        return EyesConnector(
            b_info,
            self.configure.clone(),
            deepcopy(self.server_connector),
            agent_run_id,
            job_info,
        )

    def _try_set_target_selector(self, check_settings):
        # type: (SeleniumCheckSettings) -> None
        element = check_settings.values.target_element
        if element is None:
            target_selector = check_settings.values.target_selector
            if target_selector:
                by, value = target_selector
                element = self.driver.find_element(by, value)

        if element is None:
            return None
        element = eyes_selenium_utils.get_underlying_webelement(element)
        xpath = self.driver.execute_script(GET_ELEMENT_XPATH_JS, element)
        vgs = VisualGridSelector(xpath, "target")
        check_settings.values.selector = vgs

    def get_region_xpaths(self, check_settings):
        # type: (SeleniumCheckSettings) -> List[List[VisualGridSelector]]
        element_lists = self.collect_selenium_regions(check_settings)
        frame_chain = self.driver.frame_chain.clone()
        result = []
        for element_list in element_lists:
            xpaths = []
            for elem_region in element_list:
                if elem_region.webelement is None:
                    continue
                xpath = self.driver.execute_script(
                    GET_ELEMENT_XPATH_JS, elem_region.webelement
                )
                xpaths.append(VisualGridSelector(xpath, elem_region.region_provider))
            result.append(xpaths)
        self.driver.switch_to.frames(frame_chain)
        return result

    def collect_selenium_regions(self, check_settings):
        # type: (SeleniumCheckSettings) -> List[List[WebElementRegion]]
        ignore_elements = self.get_elements_from_regions(
            check_settings.values.ignore_regions
        )
        layout_elements = self.get_elements_from_regions(
            check_settings.values.layout_regions
        )
        strict_elements = self.get_elements_from_regions(
            check_settings.values.strict_regions
        )
        content_elements = self.get_elements_from_regions(
            check_settings.values.content_regions
        )
        floating_elements = self.get_elements_from_regions(
            check_settings.values.floating_regions
        )
        accessibility_elements = self.get_elements_from_regions(
            check_settings.values.accessibility_regions
        )
        element = check_settings.values.target_element
        if element is None:
            target_selector = check_settings.values.target_selector
            if target_selector:
                by, value = target_selector
                element = self.driver.find_element(by, value)

        targets = [WebElementRegion("target", element)]
        return [
            ignore_elements,
            layout_elements,
            strict_elements,
            content_elements,
            floating_elements,
            accessibility_elements,
            targets,
        ]

    def get_elements_from_regions(self, regions_provider):
        # type:(List[GetRegion])->List[WebElementRegion]
        elements = []
        for rp in regions_provider:
            if not hasattr(rp, "get_elements"):
                continue
            webelements = rp.get_elements(self.driver)
            for elem in webelements:
                elements.append(WebElementRegion(rp, elem))
        return elements

    def _get_viewport_size(self):
        argument_guard.not_none(self.driver)
        return eyes_selenium_utils.get_viewport_size(self.driver)

    def _set_viewport_size(self):
        viewport_size = self.configure.viewport_size

        if viewport_size is None:
            for render_bi in self.configure.browsers_info:
                if isinstance(render_bi, (DesktopBrowserInfo, RenderBrowserInfo)):
                    viewport_size = render_bi.viewport_size
                    break

        if viewport_size is None:
            self.configure.viewport_size = self._get_viewport_size()
            return

        self.configure.viewport_size = viewport_size
        with self.driver.switch_to.frames_and_back([]):
            try:
                eyes_selenium_utils.set_viewport_size(self.driver, viewport_size)
            except Exception as e:
                self.logger.exception("set_viewport_size failure")
                raise_from(EyesError("Failed to set viewport size"), e)

    @staticmethod
    def _effective_disable_browser_fetching(config, check_settings):
        if check_settings.values.disable_browser_fetching is not None:
            return check_settings.values.disable_browser_fetching
        else:
            return config.disable_browser_fetching

    @staticmethod
    def _effective_layout_breakpoints(config, check_settings):
        if check_settings.values.layout_breakpoints is None:
            return config.layout_breakpoints
        else:
            return check_settings.values.layout_breakpoints


def _group_tests_by_width(running_tests, layout_breakpoints):
    # type: (List[RunningTest], Optional[List[int]]) -> Dict[int, List[RunningTest]]
    if isinstance(layout_breakpoints, list):
        layout_breakpoints = sorted(set(layout_breakpoints))
        layout_breakpoints = [layout_breakpoints[0] - 1] + layout_breakpoints

        def matching_breakpoint(test):
            i = bisect_right(layout_breakpoints, test.browser_info.width)
            return layout_breakpoints[max(i - 1, 0)]

    elif layout_breakpoints is True:

        def matching_breakpoint(test):
            return test.browser_info.width

    else:

        def matching_breakpoint(test):
            return None

    running_tests.sort(key=lambda test: test.browser_info.width)
    grouped = itertools.groupby(running_tests, matching_breakpoint)
    return {width: list(tests) for width, tests in grouped}



import itertools
import typing
import uuid
from collections import deque

import attr
from transitions import Machine

from applitools.common import RenderStatus, logger

from .vg_task import VGTask

if typing.TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Text

    from structlog import BoundLogger

    from applitools.common import (
        RenderBrowserInfo,
        RenderStatusResults,
        TestResults,
        VisualGridSelector,
    )
    from applitools.common.selenium import Configuration
    from applitools.selenium.fluent import SeleniumCheckSettings
    from applitools.selenium.visual_grid.rendreing_service import RenderingService

    from .eyes_connector import EyesConnector

NEW = "new"
NOT_OPENED = "not_opened"
OPENED = "opened"
COMPLETED = "completed"
TESTED = "tested"

STATES = [NEW, OPENED, NOT_OPENED, COMPLETED, TESTED]
TRANSITIONS = [
    {"trigger": "becomes_not_opened", "source": NEW, "dest": NOT_OPENED},
    {"trigger": "becomes_opened", "source": NOT_OPENED, "dest": OPENED},
    {
        "trigger": "becomes_tested",
        "source": [NEW, NOT_OPENED, OPENED],
        "dest": TESTED,
    },
    {
        "trigger": "becomes_completed",
        "source": [NEW, NOT_OPENED, OPENED, TESTED],
        "dest": COMPLETED,
    },
]

END_OF_CHECKS = object()


@attr.s(hash=False, str=False)
class RunningTestCheck(object):
    name = attr.ib()  # type: Text
    uuid = attr.ib(init=False, repr=False, factory=lambda: str(uuid.uuid4()))
    running_test = attr.ib(repr=False)  # type: RunningTest

    check_settings = attr.ib()  # type: SeleniumCheckSettings
    region_selectors = attr.ib()  # type: List[List[VisualGridSelector]]
    source = attr.ib()  # type: Text

    logger = attr.ib(init=False, repr=False)  # type: BoundLogger
    regions = attr.ib(init=False, factory=list)
    queue = attr.ib(init=False, factory=deque)

    def set_render_request(self, render_request):
        short_description = "{} of {}".format(
            self.running_test.configuration.test_name,
            self.running_test.configuration.app_name,
        )
        tag = self.check_settings.values.name
        render_task = self._render_task(tag, short_description, render_request)
        self.queue = deque([render_task])

    def __attrs_post_init__(self):
        self.logger = self.running_test.logger.bind(running_test_check=self)

    def __hash__(self):
        return hash(self.name + self.uuid)

    def _render_task(self, tag, short_description, render_request):
        def render_task_succeeded(render_status):
            # type: (List[RenderStatusResults]) -> None
            self.logger.debug("render_task_succeeded", task_uuid=render_task.uuid)
            if render_status:
                self.running_test.eyes.render_status_for_task(
                    render_task.uuid, render_status
                )
                if render_status.status == RenderStatus.RENDERED:
                    for vgr in render_status.selector_regions:
                        if vgr.error:
                            self.logger.error("Region error", vgr_error=vgr.error)
                        else:
                            self.regions.append(vgr.to_region())
                    self.logger.debug(
                        "render_task_succeeded rendered",
                        task_uuid=render_task.uuid,
                        regions=self.regions,
                    )
                    # schedule check task
                    self.queue.append(self._check_task(render_task, tag))
                elif render_status and render_status.status == RenderStatus.ERROR:
                    self.running_test.task_queue.clear()
                    self.running_test.open_queue.clear()
                    self.running_test.close_queue.clear()
                    self.running_test.watch_open = {}
                    self.running_test.watch_task = {}
                    self.running_test.watch_close = {}

                    self.running_test.abort()
                    self.running_test.becomes_tested()
            else:
                self.logger.error("Wrong render status", render_status=render_status)
                self.running_test.becomes_completed()

        def render_task_error(e):
            self.logger.debug(
                "render_task_error", task_uuid=render_task.uuid, exc_info=e
            )
            self.running_test.pending_exceptions.append(e)
            self.running_test.abort()
            self.running_test.becomes_tested()

        render_task = VGTask(
            "RunningTest.render {} - {}".format(short_description, tag),
            lambda: self.running_test.rendering_service.render(
                render_request, render_task_succeeded, render_task_error
            ),
            self.logger,
        )
        return render_task

    def _check_task(self, render_task, tag):
        def check_run():
            self.logger.debug("check_run", task_uuid=render_task.uuid)
            self.running_test.eyes.check(
                self.check_settings,
                render_task.uuid,
                self.region_selectors,
                self.regions,
                self.source,
            )

        check_task = VGTask(
            "perform check {} {}".format(tag, self.check_settings),
            check_run,
            self.logger,
        )

        def check_task_completed():
            # type: () -> None
            self.logger.debug("check_task_completed", task_uuid=check_task.uuid)
            if (
                self.running_test.task_lock
                and self.running_test.task_lock.uuid == self.uuid
            ):
                self.running_test.task_lock = None

            self.running_test.watch_task[self] = True
            self.running_test.maybe_becomes_tested()

        def check_task_error(e):
            self.logger.debug("check_task_error", task_uuid=check_task.uuid)
            self.running_test.pending_exceptions.append(e)
            self.running_test.becomes_completed()

        check_task.on_task_completed(check_task_completed)
        check_task.on_task_error(check_task_error)
        return check_task


@attr.s(hash=True, str=False)
class RunningTest(object):
    eyes = attr.ib(hash=False, repr=False)  # type: EyesConnector
    configuration = attr.ib(hash=False, repr=False)  # type: Configuration
    browser_info = attr.ib()  # type: RenderBrowserInfo
    rendering_service = attr.ib(hash=False, repr=False)  # type: RenderingService
    logger = attr.ib(hash=False, repr=False)  # type: BoundLogger

    tasks_list = attr.ib(init=False, factory=list, hash=False)
    test_uuid = attr.ib(init=False)
    on_results = attr.ib(init=False, hash=False)
    state = attr.ib(init=False, hash=False)

    def __str__(self):
        return "RunningTest: state={} uuid={}".format(
            self.state, format(self.test_uuid)
        )

    def __attrs_post_init__(self):
        # type: () -> None
        self.logger = self.logger.bind(running_test=self)
        self._initialize_vars()
        self._is_closed = False
        self._is_aborted = False
        self._initialize_state_machine()
        self.open()

    def _initialize_vars(self):
        # type: () -> None
        self.open_queue = deque()  # type: deque[VGTask]
        self.task_queue = deque()  # type: deque[RunningTestCheck]
        self.close_queue = deque()  # type: deque[VGTask]
        self.watch_open = {}  # type: Dict[VGTask, bool]
        self.watch_task = {}  # type: Dict[RunningTestCheck, bool]
        self.watch_close = {}  # type: Dict[VGTask, bool]
        self.task_lock = None  # type: Optional[VGTask]
        self.test_result = None  # type: Optional[TestResults]
        self.pending_exceptions = deque()  # type: deque[Exception]

    def _initialize_state_machine(self):
        # type: () -> None
        machine = Machine(
            model=self,
            states=STATES,
            transitions=TRANSITIONS,
            initial=NEW,
            send_event=True,
            queued=True,
        )
        self.machine = machine

    def on_results_received(self, func):
        self.on_results = func

    @property
    def queue(self):
        # type: () -> deque[VGTask]
        if self.state == NEW:
            return deque()
        elif self.state == NOT_OPENED:
            return self.open_queue
        elif self.state == OPENED:
            if self.task_lock:
                return self.task_lock.queue
            elif self.task_queue:
                item = self.task_queue.popleft()
                if item is END_OF_CHECKS:
                    # all checks are done and test is finished
                    return self.close_queue
                else:
                    self.task_lock = item
                    return self.task_lock.queue
            return deque()
        elif self.state == TESTED:
            return self.close_queue
        elif self.state == COMPLETED:
            return deque()
        else:
            raise TypeError("Unsupported state")

    def open(self):
        # type: () -> None
        open_task = VGTask(
            "open {}".format(self.browser_info),
            lambda: self.eyes.open(self.configuration),
            self.logger,
        )
        self.logger.debug("RunningTest VGTask created", task_name=open_task.name)

        def open_task_succeeded(test_result):
            # type: (Optional[Any]) -> None
            self.logger.debug("open_task_succeeded")
            self.watch_open[open_task] = True
            if self.all_tasks_completed(self.watch_open):
                if self.state in (TESTED, COMPLETED):
                    self.logger.debug("open_task_succeeded: test session was aborted")
                    return
                self.becomes_opened()

        def open_task_error(e):
            self.logger.debug("open_task_error", exc_info=e)
            self.pending_exceptions.append(e)
            self.becomes_completed()

        open_task.on_task_succeeded(open_task_succeeded)
        open_task.on_task_error(open_task_error)
        self.open_queue.append(open_task)
        self.watch_open[open_task] = False

    def check(
        self,
        check_settings,  # type: SeleniumCheckSettings
        region_selectors,  # type: List[List[VisualGridSelector]]
        source,  # type: Optional[Text]
    ):
        # type: (...) -> RunningTestCheck
        self.logger.debug("RunningTest.check", check_settings=check_settings)
        running_test_check_task = RunningTestCheck(
            "RunningTestCheck({})".format(check_settings.values.name),
            self,
            check_settings,
            region_selectors,
            source,
        )

        self.task_queue.append(running_test_check_task)
        self.watch_task[running_test_check_task] = False
        return running_test_check_task

    def close(self):
        # type: () -> Optional[Any]
        self._is_closed = True
        if self.state == NEW:
            self.becomes_completed()
            return None
        elif self.state in [NOT_OPENED, OPENED, TESTED]:
            close_task = VGTask(
                "close {}".format(self.browser_info),
                lambda: self.eyes.is_open and self.eyes.close(False),
                self.logger,
            )
            self.logger.debug("RunningTest VGTask", task_name=close_task.name)

            def close_task_succeeded(test_result):
                self.logger.debug("close_task_succeeded")
                # abort() could add test_result already
                if self.test_result is None and test_result:
                    self.test_result = test_result
                    if callable(self.on_results):
                        self.on_results(test=self, test_result=test_result)

            def close_task_completed():
                # type: () -> None
                self.logger.debug("close_task_completed")
                self.watch_close[close_task] = True
                if self.all_tasks_completed(self.watch_close):
                    self.becomes_completed()

            def close_task_error(e):
                self.logger.debug("close_task_error", exc_info=e)
                self.pending_exceptions.append(e)

            close_task.on_task_succeeded(close_task_succeeded)
            close_task.on_task_completed(close_task_completed)
            close_task.on_task_error(close_task_error)
            self.close_queue.append(close_task)
            self.watch_close[close_task] = False
            self.task_queue.append(END_OF_CHECKS)

    @property
    def is_closed(self):
        return self._is_closed or self._is_aborted

    def abort(self):
        if self.state == COMPLETED:
            return None
        self._is_aborted = True

        while self.close_queue:
            close_task = self.close_queue.pop()
            self.watch_close.pop(close_task)

        def ensure_and_abort():
            if not self.eyes.is_open:
                # open new session if no opened
                self.eyes._ensure_running_session()
            return self.eyes.abort()

        abort_task = VGTask(
            "abort {}".format(self.browser_info), ensure_and_abort, self.logger
        )
        self.logger.debug("RunningTest VGTask created", task_name=abort_task.name)

        def abort_task_succeeded(test_result):
            self.logger.debug("abort_task_succeeded", task_uuid=abort_task.uuid)
            self.test_result = test_result
            if callable(self.on_results):
                self.on_results(test=self, test_result=test_result)

        def abort_task_completed():
            # type: () -> None
            self.logger.debug("abort_task_completed", task_uuid=abort_task.uuid)
            self.watch_close[abort_task] = True
            if self.all_tasks_completed(self.watch_close):
                self.becomes_completed()

        def abort_task_error(e):
            self.logger.debug("abort_task_error", task_uuid=abort_task.uuid, exc_info=e)
            self.pending_exceptions.append(e)

        abort_task.on_task_succeeded(abort_task_succeeded)
        abort_task.on_task_completed(abort_task_completed)
        abort_task.on_task_error(abort_task_error)

        self.close_queue.append(abort_task)
        self.watch_close[abort_task] = False
        self.task_queue.append(END_OF_CHECKS)

    def all_tasks_completed(self, watch):
        # type: (Dict) -> bool
        if self.state == "completed":
            return True
        return all(watch.values())

    @staticmethod
    def _options_dict(configuration_options, check_settings_options):
        return {
            o.key: o.value
            for o in itertools.chain(
                configuration_options or (), check_settings_options or ()
            )
        }

    def maybe_becomes_tested(self):
        if self.close_queue and self.all_tasks_completed(self.watch_task):
            self.becomes_tested()

    @property
    def has_checks(self):
        return bool(self.watch_task)
